//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: action.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/action.h,v $
//  
//  File	      : action.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : Actions
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:35:10 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:40 1996
//  Update Count      : 500
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __ACTION_H__
#define __ACTION_H__

#include "extern.h"
#include "binding.h"
#include "exp.h"
#include "exp-fun.h"
#include "relation.h"

// Classes and types

class Action;
class GoalAction;

class Primitive;
class Post;
class Load;
class Parse;
class Achieve;
class Maintain;
class Wait;
class Query;
class Test;
class Assert;
class Retract;
class Update;

class Expression;

//
// Action
//

class Action
{

protected:

  char* _name;

  // trace information
  char* _file;
  long	_line;

  //
  
  void init(const char* name, char* file=0, long line=0)
  {
    ::strcpy(_name=new char[::strlen(name)+1], name);
  }

public:

  enum ActStatus {
    ACT_SUCCEEDED,		// Ran to completion successfully
    ACT_FAILED,			// Found impossible to finish 
    ACT_CANNOT_EXECUTE		// Not directly executable. eg. goal actions
  };

  enum ActType { 
    ACT_UNDEFINED, 
    ACT_PRIMITIVE,
    ACT_LOAD,
    ACT_PARSE,
    ACT_ASSIGN,
    ACT_FACT,
    ACT_RETRIEVE,
    ACT_TEST,
    ACT_ASSERT, 
    ACT_FAIL,
    ACT_RETRACT, 
    ACT_UPDATE,
    ACT_POST,
    ACT_UNPOST,
    ACT_GOAL_ACTION,
    ACT_ACHIEVE,
    ACT_MAINTAIN, 
    ACT_WAIT, 
    ACT_QUERY
  };


  virtual ~Action()
  {
    delete _name;
  }

  Action()
  {
    _name = 0;
    _file = 0;
    _line = 0;
  }

  Action(const char* name)
  {
    init(name);
  }

  Action(Action& a)
  {
    init(a._name);
  }

  //
  
  void set_trace(char* file, long line)
  {
    _file = file;
    _line = line;
  }

  const char* trace_file() const
  {
    return _file;
  }

  const long trace_line() const
  {
    return _line;
  }

  const char* get_name() const
  {
    return _name;
  }

  virtual Relation* get_relation() const
  {
    return 0;
  }
  
  virtual const ActType type() const
  {
    return ACT_UNDEFINED;
  }

  virtual int is_executable_action() const	= 0;

  virtual ActStatus execute(Binding* b)		= 0;
};

//
// GoalAction
//

class GoalAction : public Action
{
protected:
  
  Relation* 	_goal;
  Expression* 	_priority;

  ExpList* 	_by;
  ExpList*	_not_by;

public:

  int is_eligible(const char* name, Binding* binding);

public:

  virtual ~GoalAction()
  {
    delete _goal;
    delete _priority;

    explist_delete(_by);
    explist_delete(_not_by);
  }

  GoalAction(const char* 	name,
	     Relation* 		goal,
	     Expression* 	priority=0,
	     ExpList* 		by=0,
	     ExpList* 		not_by=0)
  : Action(name)
  {
    _goal = goal;
    _priority = priority;
    _by = by;
    _not_by = not_by;
  }
  
  GoalAction(GoalAction* g, Binding* binding)
  : Action(g->_name)
  {
    // create a new GoalAction of which relation and priority are evaluated
    _goal = new Relation(g->_goal, binding);
    _priority = new Value(g->eval_priority(binding));
    _by = explist_eval_new(g->_by, binding);
    _not_by = explist_eval_new(g->_not_by, binding);
  }

  const ActType type() const
  {
    return ACT_GOAL_ACTION;
  }
  
  Relation* get_relation() const
  {
    return _goal;
  }
  
  Expression* get_priority() const
  {
    return _priority;
  }

  double eval_priority(Binding* binding)
  {
    return _priority ? _priority->eval(binding).get_real() : 0.0;
  }

  //

  Expression* 	set_priority(Expression* priority)
  {
    return _priority = priority;
  }

  ExpList* 	set_by(ExpList* by)
  {
    return _by = by;
  }

  ExpList* 	set_not_by(ExpList* not_by)
  {
    return _not_by = not_by;
  }

  //
  
  
  int is_executable_action() const
  {
    return 0;
  }

  ActStatus execute(Binding* b)
  {
    return ACT_CANNOT_EXECUTE;
  }

public:
  
  void format_args(ostream& s, Binding* b, char* head="", char* tail="");

  virtual void format(ostream& s, Binding* b)
  {};
};


//
// Action List
//

typedef Dlist<Action*>		ActionList;
typedef DlistIterator<Action*>	ActionListIterator;


/////
/////
/////
///// Actions
/////
/////
/////


//
// Primitive
//

class Primitive : public Action
{
  Function function;
  long arity;
  ExpList* args;

public:

  ~Primitive()
  {
    args->delete_entities();
    delete args;
  }

  Primitive(char* name)
  : Action(name)
  {
    function = 0;
    arity = 0;
    args = 0;
  }

  Primitive(char* name, Function f, ExpList* el)
  : Action(name), function(f)
  {
    args = el;
    arity = el ? el->get_count() : 0;
  }

  // access functions
  long get_arity() const
  {
    return arity;
  }
  
  const ActType type() const
  {
    return ACT_PRIMITIVE;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


//
// Load
// 

class Load : public Action
{
  ExpList* args;

public:

  ~Load()
  {
    args->delete_entities();
    delete args;
  }

  Load(ExpList* el)
  : Action("LOAD")
  {
    args = el;
  }

  // access functions
  const ActType type() const
  {
    return ACT_LOAD;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};

//
//
//

class Parse : public Action
{
  ExpList* args;

public:

  ~Parse()
  {
    args->delete_entities();
    delete args;
  }

  Parse(ExpList* el)
  : Action("PARSE")
  {
    args = el;
  }

  // access functions
  const ActType type() const
  {
    return ACT_PARSE;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};

//
// Assign
//

class Assign : public Action
{
  Expression* var;
  Expression* exp;

public:

  ~Assign()
  {
    delete var;
    delete exp;
  }

  Assign(Expression* v, Expression* e)
  : Action(v->get_name())
  {
    var = v;
    exp = e;
  }

  // access functions
  const ActType type() const
  {
    return ACT_ASSIGN;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


//
// Fact
//

class Fact : public Action
{
  Relation* relation;

public:

  ~Fact()
  {
    delete relation;
  }

  Fact(Relation* r)
  : Action(r->get_name())
  {
    relation = r;
  }

  // access functions
  const ActType type() const
  {
    return ACT_FACT;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


//
// Fail
//

class Fail : public Action
{
  
public:

  const ActType type() const
  {
    return ACT_FAIL;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};

//
// Retrieve
//
//

/*
   This action is simliar to 'FACT' action, but it gets new values from the
   WM for the variables instead of using current value of the variable for
   the matching: 
  
   NOTE:
	Once this action is executed and it fails to get new values,
	the variables occurring in the arguments will LOSE their values.
*/

class Retrieve : public Action
{
  Relation* relation;

public:

  ~Retrieve()
  {
    delete relation;
  }

  Retrieve(Relation* r)
  : Action(r->get_name())
  {
    relation = r;
  }

  // access functions
  const ActType type() const
  {
    return ACT_RETRIEVE;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


//
// Test
//

class Test : public Action
{
  Expression* exp;
  
public:

  ~Test()
  {
    delete exp;
  }

  Test(Expression* e)
  : Action(e->get_name())
  {
    exp = e;
  }
  
  // access functions
  Expression* get_exp() const
  {
    return exp;
  }
  
  const ActType type() const
  {
    return ACT_TEST;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};

/////
/////
/////
/////
/////
/////
/////

//
// Assert
//

class Assert : public Action
{
  Relation* relation;
  
public:

  ~Assert()
  {
    delete relation;
  }

  Assert(Relation* rel)
  : Action(rel->get_name())
  {
    relation = rel;
  }
  
  // access functions
  Relation* get_relation() const
  {
    return relation;
  }

  const ActType type() const
  {
    return ACT_ASSERT;
  }

  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


//
// Retract
//

class Retract : public Action
{
  Relation* relation;
  
public:

  ~Retract()
  {
    delete relation;
  }

  Retract(Relation* rel)
  : Action(rel->get_name())
  {
    relation = rel;
  }
  
  // access functions
  Relation* get_relation() const
  {
    return relation;
  }
  
  const ActType type() const
  {
    return ACT_RETRACT;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


//
// Update
//

class Update : public Action
{
  Relation* old_relation;
  Relation* new_relation;
  
public:

  ~Update()
  {
    delete old_relation;
    delete new_relation;
  }

  Update(Relation* old_rel, Relation* new_rel)
  : Action(old_rel->get_name())
  {
    old_relation = old_rel;
    new_relation = new_rel;
  }
  
  // access functions
  Relation*	get_old_relation() const
  {
    return old_relation;
  }
  
  Relation* get_new_relation() const
  {
    return new_relation;
  }
  
  const ActType type() const
  {
    return ACT_UPDATE;
  }
  
  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


class Post : public Action
{
  GoalAction* 	_goal_action;

public:

  ~Post()
  {
    delete _goal_action;
  };

  Post(GoalAction* goal_action)
  : Action("POST")
  {
    _goal_action = goal_action;
  }
  
  const ActType type() const
  {
    return ACT_POST;
  }

  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};


class UnPost : public Action
{
  GoalAction* 	_goal_action;

public:

  ~UnPost()
  {
    delete _goal_action;
  };

  UnPost(GoalAction* goal_action)
  : Action("UNPOST")
  {
    _goal_action = goal_action;
  }
  
  const ActType type() const
  {
    return ACT_UNPOST;
  }

  int is_executable_action() const
  {
    return 1;
  }

  ActStatus execute(Binding* b);
};

/////
/////
/////
///// Goal Actions
/////
/////
/////

class Achieve : public GoalAction
{
public:

  ~Achieve()
  {};

  Achieve(const char* 	name,
	  Relation* 	goal,
	  Expression* 	priority=0,
	  ExpList* 	by=0,
	  ExpList* 	not_by=0)
  : GoalAction(name, goal, priority, by, not_by)
  {};
  
  const ActType type() const
  {
    return ACT_ACHIEVE;
  }

  void format(ostream& s, Binding* b)
  {
    format_args(s, b, "ACHIEVE ", ";");
  }
};


//
// Maintain
//

class Maintain : public GoalAction
{

public:

  ~Maintain()
  {};

  Maintain(char* name, Relation* goal, Expression* priority=0)
  : GoalAction(name, goal, priority)
  {};
  
  const ActType type() const
  {
    return ACT_MAINTAIN;
  }

  void format(ostream& s, Binding* b)
  {
    format_args(s, b, "MAINTAIN ", ";");
  }
};


//
// Wait
//

class Wait : public GoalAction
{
  
public:

  ~Wait()
  {};

  Wait(char* name, Relation* goal, Expression* priority=0)
  : GoalAction(name, goal, priority)
  {};
  
  const ActType type() const
  {
    return ACT_WAIT;
  }

  void format(ostream& s, Binding* b)
  {
    format_args(s, b, "WAIT ", ";");
  }
};


//
// Query
//

class Query : public GoalAction
{
  Expression* _test;
  
public:

  ~Query()
  {
    delete _test;
  }

  Query(Relation* rel, Expression* test=0, Expression* priority=0)
  : GoalAction(rel->get_name(), rel, priority)
  {
    _test = test;
  }

  //
  
  Expression* 	set_test(Expression* test)
  {
    return _test = test;
  }

  //
  
  // access functions
  const ActType type() const
  {
    return ACT_QUERY;
  }
  
  Expression* get_test() const
  {
    return _test;
  }

  void format(ostream& s, Binding* b)
  {
    format_args(s, b, "MAINTAIN ", ";");
  }
};

#endif




