//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: goal.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/goal.cc,v $
//  
//  File	      : goal.h
//  Author(s)	      : Marcus J. Huber <marcush@engin.umich.edu>
//			Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : Goal Structure
//  
//  Original author(s): Marcus J. Huber <marcush@engin.umich.edu>
//			Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:29:21 1993
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:49:53 1997
//  Update Count      : 234
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stream.h>
#include <strstream.h>
#include <strings.h>

#include "extern.h"
#include "relation.h"
#include "goal.h"
#include "ka.h"


////
////
//// Goal
////
////

/////////////////////////////////////////////////////////////////////////////
//
int Goal::generate_soak()
{
  if (
      // If the goal hasnt been completed yet and hasn't been UNPOSTed
      (status != IS_SUCCESS && status != IS_ABANDONED) &&
      
      // and the goal is a "leaf" goal.
      (subgoal == 0)) {

    // If it is "new" (meaning that it has just been POSTed, subgoaled,
    // or has been refreshed by the intention structure) then definitely
    // generate a SOAK

    if (new_goal) return 1;

    // If the goal is not new, but does not yet have an intended KA
    if (!intention) {

      // If anything in the world model has changed then we need to
      // consider the goal for SOAK generation
      if (program.world_model->any_new()) return 1;
    }
  }
  return 0;
}

/////////////////////////////////////////////////////////////////////////////
//
// Check to see if the stack in which this goal is part is blocked.
//
int Goal::is_stack_blocked()
{
  Goal*		gl = prev_goal;

  // Check this goal
  if (status == IS_BLOCKED)
    return 1;

  // First move up to the top of the stack for this goal (if any), checking
  // on the way up.
  while (gl) {
    if (gl->get_status() == IS_BLOCKED)
      return 1;
    gl = gl->get_prev_goal();
  }

  // Then check from the goal down
  gl = subgoal;
  while (gl) {
    if (gl->get_status() == IS_BLOCKED)
      return 1;
    gl = gl->get_subgoal();
  }

  return 0;

}

/////////////////////////////////////////////////////////////////////////////
//
// If the goal does not match the pattern relation return 0
// otherwise return new binding for the pattern binding
//
int Goal::match_relation(Relation* patt_relation, Binding* patt_binding)
{
  return unify(patt_relation, patt_binding, get_relation(), get_goal_binding());
}


/////////////////////////////////////////////////////////////////////////////
//
int Goal::match_goal(GoalAction* goal_action, Binding* goal_action_binding)
{
  if (match_relation(goal_action->get_relation(), goal_action_binding))
    if (!goal_action->get_priority() ||
	eval_priority() == goal_action->eval_priority(goal_action_binding))
      return 1;

  return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
void Goal::print()
{
  cout << "  Goal: \t: *";
  goal->format_args(cout, get_goal_binding());
  cout << "*\n";

#ifdef UM_PRS
  
  if (intention)
    cout << form("  Priority\t: %5.2f \n", intention->eval_priority());

  cout << "  New ?    \t: " << (new_goal ? "True" : "False") << "\n";

  cout << "  Status   \t: " << status_desc[get_status()] << "\n";

  cout << "  Subgoal  \t: " << subgoal;
  cout << "=> " << (subgoal? subgoal->get_name() : "NIL") << "\n";

  cout << "  Prev_Goal\t: " << prev_goal;
  cout << "=> " << (prev_goal ? prev_goal->get_name() : "NIL") << "\n";

#endif  

}


/////////////////////////////////////////////////////////////////////////////
//
SoakElement* Goal::set_intention(SoakElement* intent)
// Try THIS means of achieving the goal
{
  if (intention)
    delete intention;
  intention = intent;
  return intention;
}


/////////////////////////////////////////////////////////////////////////////
//
// Removes the entire subgoal stack, including removing subgoals from the
// GoalSet and executing the FAILURE sections of each subgoal
//
void Goal::remove_intention(int failed)
{
  Goal*			a_goal;
  KaBodyAtomicElement*	failure_section;

  a_goal = this;

  // Move down to bottom of subgoal string
  while (a_goal && a_goal->get_subgoal()) {
    a_goal = a_goal->get_subgoal();
  }

  // Now backtrack, cleaning things up on the way
  while (a_goal && (a_goal != this)) {

    if (failed && (a_goal->get_intention())) {
      // Execute FAILURE section
      if ((failure_section =
	   a_goal->get_intention()->_ka->get_failure()) != 0) {
	a_goal->set_runtime_frame(failure_section->new_runtime_frame());
	a_goal->get_runtime_frame()->execute(a_goal->get_intention_binding(), a_goal);
      }
    }
    
    // clear up memory
    a_goal->set_intention(0);

    // Remove goal from GoalSet
    program.goal_list->remove(a_goal);
    
    a_goal = a_goal->get_prev_goal();
  }

  // Execute FAILURE section for this KA too
  if (failed) {
    if (this->get_intention() &&
	(failure_section = this->get_intention()->_ka->get_failure()) != 0) {
      this->set_runtime_frame(failure_section->new_runtime_frame());
      this->get_runtime_frame()->execute(a_goal->get_intention_binding(), this);
    }
  }

  set_intention(0);
  set_subgoal(0);
  set_runtime_frame(0);
}


////
////
//// GoalSet
////
////

/////////////////////////////////////////////////////////////////////////////
//
Goal* GoalSet::get_goal(long nth) const
{
  GoalListIterator	next_goal(_goals);
  Goal*			nth_goal;

  if (nth <= _goals->get_count()-1) {
    nth_goal = next_goal();
    for (long i = 1;i <= nth; i++) {
      nth_goal = next_goal();
    }
    return nth_goal;
  }
  return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
void GoalSet::renew_leaf_goals(void)
{
  GoalListIterator	next_goal(_goals);
  Goal*			nth_goal;
  
  nth_goal = next_goal();
  while (nth_goal) {
    if (nth_goal->is_leaf_goal())
      nth_goal->set_new();
    nth_goal = next_goal();
  }
}

/////////////////////////////////////////////////////////////////////////////
//
int GoalSet::all_goals_done(void)
{
  GoalListIterator	next_goal(_goals);
  Goal*			nth_goal;
  
  nth_goal = next_goal();
  while (nth_goal) {
    if (nth_goal->is_toplevel_goal() &&
	(nth_goal->get_status() != IS_SUCCESS))
      return 0;
    nth_goal = next_goal();
  }
  return 1;
}

/////////////////////////////////////////////////////////////////////////////
//
// Extract the given goal from the set of goals
//
void GoalSet::remove(Goal* gl)
{
  // Find the goal in the goal set and remove it
  GoalListIterator	next_goal(program.goal_list->get_goals());
  Goal*			a_goal;
  
  a_goal = next_goal();
  while (a_goal && gl != a_goal)
    a_goal = next_goal();
  
  if (a_goal) {
    next_goal.remove_this();
    if (a_goal->get_prev_goal())
      a_goal->get_prev_goal()->set_subgoal(0);
    delete a_goal;
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// Remove the given goal from the list of goals if the goal is not currently
// being executed, or flag it as dropped if it is currently executed so that
// the interpreter can remove the entire subgoal stack.
//
void GoalSet::drop(GoalAction* ga, Binding* b)
{
  // Find goal in goal list and remove it
  GoalListIterator	next_goal(program.goal_list->get_goals());
  Goal*			a_goal = next_goal();

  while (a_goal && !a_goal->match_goal(ga, b))
    a_goal = next_goal();
  
  if (a_goal) {

#ifdef UM_PRS
    // If we've found a matching goal, check to see if this is on
    // the intention structure (i.e. currently executing)
    if (a_goal->get_intention_structure()) {
      a_goal->set_status(IS_ABANDONED);
      return;
    }
#endif
    
    // Goal not being executed, so take it off the goal list
    next_goal.remove_this();
    delete a_goal;
  }
}


/////////////////////////////////////////////////////////////////////////////
//
Goal* GoalSet::add(GoalAction* goal_action, Goal* prev_goal)
{
  Goal* new_goal = new Goal(goal_action, prev_goal);
  _goals->append(new_goal);
  return new_goal;
}


/////////////////////////////////////////////////////////////////////////////
//
Goal* GoalSet::add_unique(GoalAction* goal_action, Goal* prev_goal, Binding* b)
{
  GoalListIterator	next_goal(program.goal_list->get_goals());
  Goal*			goal = next_goal();

  while (goal && !goal->match_goal(goal_action, b))
    goal = next_goal();
  
  if (goal)
    return goal;
  else
    return add(goal_action, prev_goal);
}


/////////////////////////////////////////////////////////////////////////////
//
void GoalSet::format(ostream& s, Binding* b)
{
  GoalListIterator	next_goal(_goals);
  Goal*			g;

  while(g=next_goal()) {
    g->get_goal_action()->format(s, b);
    s << "\n";
  }
}


/////////////////////////////////////////////////////////////////////////////
//
void GoalSet::print()
{
  GoalListIterator	next_goal(_goals);
  Goal*			g;
  int 			i = 1;

  while (g=next_goal()) {
    cout << "\n" << i++ << ":\n";
    g->print();
  }
}


