//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: ka.h,v 3.1.1.2 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/ka.h,v $
//  
//  File	      : ka.h
//  Author(s)	      : Marcus J. Huber <marcush@engin.umich.edu>
//
//  Description	      : Knowledge Area
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:24:09 1993
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:53:46 1997
//  Update Count      : 477
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __KA_H__
#define __KA_H__

#include "binding.h"
#include "symtab.h"
#include "action.h"
#include "condition.h"
#include "exp.h"
#include "dlist.h"
#include "is.h"

//
// forward declarations
//

class KaContext;
class KaBody;
class KaBodyElement;
class KaBodyAtomicElement;

class KaRuntimeFrame;
class KaSimpleRuntimeFrame;
class KaBranchRuntimeFrame;
class KaIfRuntimeFrame;
class KaWhileRuntimeFrame;
class KaDoRuntimeFrame;
class KaAtomicRuntimeFrame;

//
// KA Context
// 

class KaContext
{
  ConditionList* _conditions;

public:

  ~KaContext()
    {
      _conditions->delete_entities();
      delete _conditions;
    }
  
  KaContext(ConditionList* conditions)
    {
      _conditions = conditions;
    }

public:

  ConditionList* add_conditions(ConditionList* cl)
    {
      _conditions->append(*cl);
      delete cl;
      return _conditions;
    }

  ConditionList* get_conditions() const
    {
      return _conditions;
    }

  int	check(BindingList& bl) const;
  int	confirm(Binding* b) const;
};


//
// KA
//

class Ka
{
  SymbolTable*		_symtab;
  
  char*			_name;
  char*			_documentation;
  int			_valid;

  Action*		_purpose;
  KaContext*		_context;
  Expression* 		_priority;
  KaBody*		_effect;
  KaBodyAtomicElement*	_failure;
  KaBody*		_body;
  
public:
  
  ~Ka();

  Ka()
    {
      _symtab 		= new SymbolTable();
      _name 		= 0;
      _documentation 	= 0;
      _valid 		= 1;
      _purpose 		= 0;
      _context 		= 0;
      _priority 		= 0;
      _effect 		= 0;
      _failure 		= 0;
      _body 		= 0;
    }

public:

  SymbolTable*		get_symtab() const	{ return _symtab; }

  const char*		get_name() const 	{ return _name; }
  const char*		get_doc() const  	{ return _documentation; }

  int			is_valid()		{ return _valid; }
  void			disable()		{ _valid = 0; }
  
  KaContext*		get_context() const	{ return _context; }
  Action*		get_purpose() const	{ return _purpose; }
  KaBody*		get_body() const	{ return _body; }

  KaBody*		get_effect() const	{ return _effect; }
  KaBodyAtomicElement*	get_failure() const	{ return _failure; }


public:
  
  const char*	set_name(const char* ka_name)
    {
      delete _name;
      ::strcpy(_name=new char[::strlen(ka_name)+1], ka_name);
      return _name;
    }

  const char*	set_doc(const char* ka_doc)
    {
      delete _documentation;
      ::strcpy(_documentation=new char[::strlen(ka_doc)+1], ka_doc);
      return _documentation;
    }

  KaContext*	add_context(ConditionList* cl)
    {
      if (_context)
	_context->add_conditions(cl);
      else
	_context = new KaContext(cl);

      return _context;
    }

  Action*	set_purpose(Action* p)		{ return _purpose = p; }
  KaContext*	set_context(KaContext* c)	{ return _context = c; }
  Expression*	set_priority(Expression* p)	{ return _priority = p; }
  KaBody*	set_body(KaBody* b)		{ return _body = b; }

  KaBody*	set_effect(KaBody* e)
    {
      return _effect = e;
    }
  
  KaBodyAtomicElement*	set_failure(KaBodyAtomicElement* f)
    {
      return _failure = f;
    }


public:

  int check_context(BindingList& bl) const;
  int confirm_context(Binding* b) const;

  double eval_priority(Binding* binding) const
    {
      return _priority ? _priority->eval(binding).get_real() : 0.0;
    }
};


//
//
//

enum BranchType
{
  NO_BRANCH,
  AND_BRANCH,
  OR_BRANCH
};

//

class KaBodyElement
{

public:

  virtual ~KaBodyElement() {};

  enum KaElementType
    { 
      KA_UNDEFINED,
      KA_SIMPLE,
      KA_BRANCH,
      KA_WHEN,
      KA_WHILE,
      KA_DO,
      KA_ATOMIC
      };

  enum KaReturnType
    { 
      KA_ELEMENT_INCOMP,
      KA_ELEMENT_COMPLETE,
      KA_ELEMENT_FAILED
      };

  virtual int			is_final_element() const
    {
      return next_element ? 0 : 1;
    }

  virtual KaRuntimeFrame*	new_runtime_frame() = 0;

  // access functions
  virtual KaBodyElement*	get_next_element() const
    {
      return next_element;
    }
  
  virtual void			set_next_element(KaBodyElement *ne)
    {
      next_element = ne;
    }
  
  virtual KaElementType		type() const
    {
      return element_type;
    }

  virtual void			set_type(KaElementType t)
    {
      element_type = t;
    }

protected:

  KaBodyElement*	next_element;
  KaElementType		element_type;
  
  KaBodyElement()
    {
      element_type = KA_UNDEFINED;
      next_element = 0;
    }

  KaBodyElement(KaBodyElement *ne)
    {
      next_element = ne;
    }

};

//
//
// KaBodySimpleElement
//
//

class KaBodySimpleElement : public KaBodyElement
{
  Action*	action;
  
public:

  ~KaBodySimpleElement()
    {
      delete action;
    }
  
  KaBodySimpleElement(Action* a = 0)
    {
      action = a;
      element_type = KA_SIMPLE;
    }

  KaRuntimeFrame*		new_runtime_frame();

  // access functions
  Action*			get_action() const	{ return action; }
  Action*			set_action(Action *a)	{ return action = a; }

  KaElementType			type() const		{ return element_type; }
};

//
//
// KaBodyBranchElement
//
//

class KaBodyBranchElement : public KaBodyElement
{
  BranchType		branch_type;
  long			num_branches;
  KaBodyElement**	branch_start;		// array of Element pointers

public:

  ~KaBodyBranchElement();

  KaBodyBranchElement()
    {
      branch_type = NO_BRANCH;
      num_branches = 0;
      branch_start = 0;
      element_type = KA_BRANCH;
    }

  KaRuntimeFrame*	new_runtime_frame();

  void		add_branch(KaBodyElement* b);

  // access functions
  long		get_num_branches() const	{ return num_branches; }

  BranchType	get_branch_type() const		{ return branch_type; }
  void		set_branch_type(BranchType bt)	{ branch_type = bt; }

  KaBodyElement* get_branch_start(long branchnum) const
    {
      return (branchnum >= 0 && branchnum < num_branches) ? branch_start[branchnum] : 0;
    }

  KaElementType type() const			{ return element_type; }
};


//
//
// KaBodyWhenElement
//
//
class KaBodyWhenElement : public KaBodyElement
{
  Action*		test;			// The predicate 
  KaBodyElement*	start_element;		// The start of the body

public:

  ~KaBodyWhenElement();

  KaBodyWhenElement(Action* a, KaBodyElement* b)
    {
      test = a;
      start_element = b;
      element_type = KA_WHEN;
    }

  KaRuntimeFrame*	new_runtime_frame();

  // access functions
  Action*		get_test() const		{ return test; }
  Action*		set_test(Action* a)		{ return test = a; }

  KaBodyElement* 	get_start_element() const	{ return start_element; }

  KaElementType 	type() const			{ return element_type; }

};


//
//
// KaBodyWhileElement
//
//

class KaBodyWhileElement : public KaBodyElement
{
  Action*		test;			// The predicate for the loop
  KaBodyElement*	start_element;		// The start of the WHILE loop

public:

  ~KaBodyWhileElement();

  KaBodyWhileElement(Action* a, KaBodyElement* b)
    {
      test = a;
      start_element = b;
      element_type = KA_WHILE;
    }

  KaRuntimeFrame*	new_runtime_frame();

  // access functions
  Action*		get_test() const		{ return test; }
  Action*		set_test(Action* a)		{ return test = a; }

  KaBodyElement* 	get_start_element() const	{ return start_element; }

  KaElementType 	type() const			{ return element_type; }

};


//
//
// KaBodyDoElement
//
//

class KaBodyDoElement : public KaBodyElement
{
  Action*		test;			// The predicate for the loop
  KaBodyElement*	start_element;		// The start of the DO loop

public:

  ~KaBodyDoElement();

  KaBodyDoElement(Action* a, KaBodyElement* b)
    {
      test = a;
      start_element = b;
      element_type = KA_DO;
    }

  KaRuntimeFrame*	new_runtime_frame();

  // access functions
  Action*		get_test() const		{ return test; }
  Action*		set_test(Action* a)		{ return test = a; }

  KaBodyElement* 	get_start_element() const	{ return start_element; }

  KaElementType 	type() const  			{ return element_type; }

};


//
//
// KaBodyAtomicElement
//
//

class KaBodyAtomicElement : public KaBodyElement
{
  KaBodyElement*	start_element;		// The start of the body

public:

  ~KaBodyAtomicElement();

  KaBodyAtomicElement(KaBodyElement* body)
    {
      start_element = body;
      element_type = KA_ATOMIC;
    }

  KaRuntimeFrame*	new_runtime_frame();

  // access functions
  KaBodyElement* 	get_start_element() const	{ return start_element; }

  KaElementType 	type() const			{ return element_type; }

};


//
//
// KaBody
//
//

class KaBody
{
  KaBodyElement*	start_element;

public:

  ~KaBody();

  KaBody()
    {
      start_element = 0;
    }

  KaBody(KaBodyElement* be)
    {
      start_element = be;
    }

  // access functions
  KaBodyElement*	get_start_element() const
    {
      return start_element;
    }
  
  KaBodyElement*	set_start_element(KaBodyElement* be)
    {
      return start_element = be;
    }
};


//
//
// KA Table
//
//

#define KA_TAB_SIZE		128 /* Initial size to allocate for symbol table. */	
#define KA_HASHTAB_SIZE		63  /* */
#define KA_TAB_INCREMENT	128 /* Increment when symbol table must grow. */	

//

class KaEnt : public Symbol
{
  Ka* ka;			// The content of the KA Table

public:

  ~KaEnt()
    {
      delete ka;
    }
  
  KaEnt(Ka* k)
    : Symbol(k->get_purpose()->get_name())
      {
	ka = k;
      }
  
  KaEnt(KaEnt& ka_ent)
    : Symbol(ka_ent)
      {
	ka = ka_ent.ka;
      }

  //
  
  void operator=(KaEnt& ka_ent)
    {
      Symbol::operator=(ka_ent);
      ka = ka_ent.ka;
    }

  Ka* get_ka() const
    {
      return ka;
    }
};


//
//
//

class KaTable : public SymbolTable
{ 
  // KA is hashed on the 'goal-name' of the KA (not the KA name)
friend class KaTableIterator;

public:

  KaTable(long katab_sz		= KA_TAB_SIZE, 
	  long katab_inc_sz	= KA_TAB_INCREMENT,
	  long hashtab_sz	= KA_HASHTAB_SIZE)
    : SymbolTable(katab_sz, katab_inc_sz, hashtab_sz)
      {};

  //

  KaEnt* lookup(long id) const
    {
      return (KaEnt*) SymbolTable::lookup(id);
    }
  
  KaEnt* lookup(const char* name) const
    {
      return (KaEnt*) SymbolTable::lookup(name);
    }

  KaEnt* add(KaEnt* ka_ent)
    {
      return (KaEnt*) SymbolTable::add(ka_ent);
    }

  void disable(const char* name);
};


class KaTableIterator
{
  KaTable* 	_table;
  long 		_id;
  long 		_max;

public:

  KaTableIterator(KaTable* kt)
    {
      _table = kt;
      _id = 0;
      _max = _table->_nextid - 1;
    }
  
  Ka* operator()()
    {
      return (_id <= _max) ? _table->lookup(_id++)->get_ka() : 0;
    }
};

/*
   class KaTableBucketIterator
   {
   KaTable*		_table;
   const char* 		_name;
   SymbolListIterator* 	_bucket;
   
   public:

   ~KaTableBucketIterator()
   {
   delete _bucket;
   }

   KaTableBucketIterator(KaTable* kt, const char* name) 
   {
   _table = kt;
   _name = name;
   _bucket = new SymbolListIterator(kt->get_bucket(name));
   }

   public:
   
   Ka* operator()()
   {	
   KaEnt* e = (KaEnt*) _bucket->operator()();

   if (!e) return 0;
   
   if (strcmp(_name, e->get_name()) == 0)
   return e->get_ka();
   else
   return operator()();
   }

   void remove_this()
   {
   _table->remove_this(_bucket);
   }
   };
   */

class KaTableBucketIterator
{
  KaTable*		_table;
  Goal*			_goal;
  SymbolListIterator* 	_bucket;
  
public:

  ~KaTableBucketIterator()
    {
      delete _bucket;
    }

  KaTableBucketIterator(KaTable* kt, Goal* goal);

public:
  
  Ka* operator()(Binding*& ka_binding);

  void remove_this()
    {
      _table->remove_this(_bucket);
    }
};

#endif

