//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: run.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/run.h,v $
//  
//  File	      : run.cc
//  Author(s)	      : Marcus J. Huber <marcush@engin.umich.edu>
//
//  Description	      : Run-time
//  
//  Original author(s): Marcus J. Huber <marcush@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:24:09 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:03:38 1996
//  Update Count      : 163
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __RUN_H__
#define __RUN_H__

#include "dlist.h"
#include "binding.h"
#include "symtab.h"
#include "condition.h"
#include "exp.h"
#include "action.h"
#include "ka.h"


// forward declarations

class Goal;

class SoakElement;

class KaBodyElement;
class KaBodySimpleElement;
class KaBodyBranchElement;
class KaBodyWhenElement;
class KaBodyWhileElement;
class KaBodyDoElement;
class KaBodyAtomicElement;

class KaRuntimeFrame;
class KaSimpleRuntimeFrame;
class KaBranchRuntimeFrame;
class KaWhenRuntimeFrame;
class KaWhileRuntimeFrame;
class KaDoRuntimeFrame;
class KaAtomicRuntimeFrame;

//
//
// KaRuntimeFrame
//
//

class KaRuntimeFrame
{

protected:

  KaBodyElement*	this_element;
  KaRuntimeFrame*	subframe;
      
public:
  
  virtual ~KaRuntimeFrame()
  {
    if (subframe)
      delete subframe;
  }

  virtual KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal) = 0;
  virtual void				intend(SoakElement* s) = 0;

  // access functions
  virtual KaRuntimeFrame*	get_subframe() const	{ return subframe; }
  virtual void			set_subframe(KaRuntimeFrame* f)
  {
    if (subframe != 0) {
      delete subframe;
    }
    subframe = f;
  }

  virtual KaBodyElement*	get_this_element() const	{ return this_element; }
  virtual void			set_this_element(KaBodyElement* se) { this_element = se; }
};


//
//
// KaSimpleRuntimeFrame
//
//

class KaSimpleRuntimeFrame : public KaRuntimeFrame
{

public:

  KaSimpleRuntimeFrame(KaBodySimpleElement* se)	{ this_element = se;
						  subframe = 0;
						}

  KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal);
  void				intend(SoakElement* s) {};

};


//
//
// KaGoalRuntimeFrame
//
//

class KaGoalRuntimeFrame : public KaRuntimeFrame
{
  Goal*		subgoal;
  
public:

  KaGoalRuntimeFrame(KaBodySimpleElement* se)
  {
    this_element = se;
    subframe = 0;		// It's a primitive action
    subgoal = 0;
  }

  KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal);
  void				intend(SoakElement* s);

  // access functions
  Goal*				get_subgoal() const	{ return subgoal; }
  void				set_subgoal(Goal* g)	{ subgoal = g; }

};


//
//
// KaBranchRuntimeFrame
//
//

class KaBranchRuntimeFrame : public KaRuntimeFrame
{
  long			active_branch_num;	// index (0 based) to current branch
  KaBodyElement**	branch_current;		// array of Element pointers
  
public:
  KaBranchRuntimeFrame(KaBodyBranchElement* be)
  {
    this_element = be;
    subframe = be->get_branch_start(0)->new_runtime_frame();
    active_branch_num = 0;
    branch_current = new KaBodyElement*[be->get_num_branches()];
    for (int bl = 0; bl < be->get_num_branches(); bl++) {
      branch_current[bl] = be->get_branch_start(bl);
    }
  }

  ~KaBranchRuntimeFrame()				{ delete branch_current; }
  
  KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal);
  void				intend(SoakElement* s)	{ if (subframe)
							    subframe->intend(s);
							}

  // access functions
  KaBodyElement*	get_branch_current(long branchnum) const;
  void			set_branch_current(long branchnum, KaBodyElement* re);

  long			get_active_branch_num()	const	{ return active_branch_num; }
  void			set_active_branch_num(long n);
};


//
//
// KaWhenRuntimeFrame
//
//

class KaWhenRuntimeFrame : public KaRuntimeFrame
{
  KaBodyElement*	current_element;
  
public:

  KaWhenRuntimeFrame(KaBodyWhenElement* we)
  {
    this_element = we;
    subframe = we->get_start_element()->new_runtime_frame();
    current_element = we->get_start_element();
  }

  KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal);
  void				intend(SoakElement* s)	{ if (subframe)
							    subframe->intend(s);
							}

  // access functions
  KaBodyElement*	get_current_element() const	{ return current_element; }
  void			set_current_element(KaBodyElement* e) { current_element = e; }
};


//
//
// KaWhileRuntimeFrame
//
//

class KaWhileRuntimeFrame : public KaRuntimeFrame
{
  KaBodyElement*	current_element;
  int			check_loop_cond;
  
public:

  KaWhileRuntimeFrame(KaBodyWhileElement* we)
  {
    this_element = we;
    subframe = we->get_start_element()->new_runtime_frame();
    current_element = we->get_start_element();
    check_loop_cond = 1;
  }

  KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal);
  void				intend(SoakElement* s)	{ if (subframe)
							    subframe->intend(s);
							}

  // access functions
  KaBodyElement*	get_current_element() const	{ return current_element; }
  void			set_current_element(KaBodyElement* e) { current_element = e; }
};


//
//
// KaDoRuntimeFrame
//
//

class KaDoRuntimeFrame : public KaRuntimeFrame
{
  KaBodyElement*	current_element;
  
public:

  KaDoRuntimeFrame(KaBodyDoElement* we)
  {
    this_element = we;
    subframe = we->get_start_element()->new_runtime_frame();
    current_element = we->get_start_element();
  }

  KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal);
  void				intend(SoakElement* s)	{ if (subframe)
							    subframe->intend(s);
							}


  // access functions
  KaBodyElement*	get_current_element() const	{ return current_element; }
  void			set_current_element(KaBodyElement* e) { current_element = e; }
};


//
//
// KaAtomicRuntimeFrame
//
//

class KaAtomicRuntimeFrame : public KaRuntimeFrame
{
  // Probably don't need this since we don't need to remember
  // location between interpreter cycles.
  KaBodyElement*	current_element;
  
public:

  KaAtomicRuntimeFrame(KaBodyAtomicElement* we)
  {
    this_element = we;
    subframe = we->get_start_element()->new_runtime_frame();
    current_element = we->get_start_element();
  }

  KaBodyElement::KaReturnType	execute(Binding* b, Goal* this_goal);
  void				intend(SoakElement* s)  { if (subframe)
							    subframe->intend(s);
							}

  // access functions
  KaBodyElement*	get_current_element() const		{ return current_element; }
  void			set_current_element(KaBodyElement* e)	{ current_element = e; }
};




#endif

