//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: soak.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/soak.cc,v $
//  
//  File	      : soak.cc
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  		      : Marcus J. Huber <marcush@engin.umich.edu>
//
//  Description	      : Set of Applicable Knowledge-area
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:21:36 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:53 1996
//  Update Count      : 251
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>
#include <values.h>

#include "extern.h"
#include "binding.h"
#include "relation.h"
#include "symtab.h"
#include "wm.h"
#include "ka.h"
#include "goal.h"
#include "random.h"
#include "soak.h"

//
// SoakElement
//
double SoakElement::eval_priority()
{
  // Note that goal priority uses its own internal binding and
  // KA uses the SE binding

  double goal_priority = _from_goal->eval_priority();
  double ka_priority = _ka->eval_priority(_binding);

  if (program.priority_evaluation_function_hook)
    return program.priority_evaluation_function_hook(goal_priority, ka_priority);
  else
    return goal_priority + ka_priority;
}


void SoakElement::print(ostream& s)
{ 
  s << "KA Name: " << _ka->get_name() << "\n";
  s << "Binding:\n";
  _binding->print(s);
}


//
// Soak
//
SoakElement* Soak::add(Ka* ka, Goal* g, Binding* b)
{
  SoakElement*	se = new SoakElement;
  se->_ka = ka;
  se->_from_goal = g;
  se->_binding = b;

  _elements->append(se);

  return se;
}


inline long Soak::random_uniform(long range) const
{
  static Random_Uniform_Integer ru(MAXLONG);

  return ru.generate() % range;
}


void Soak::instantiate(Ka* ka, Binding* ka_binding, Goal* goal)
// check only the purpose and the context of the KA
{
  BindingList bl(ka_binding);

  if (ka->check_context(bl)) {

    BindingListIterator next_binding(bl);
    Binding* b;

    while(b = next_binding()) {
      if (goal->is_new() || b->is_new_wm_binding())  
	add(ka, goal, b);
      else
	delete b;
    }
  }
}


Soak::Soak(KaTable* kt, WmTable* wt, GoalSet* gt, int test_generate)
{
  register Ka*		ka;
  register Relation*	wm;
  register Goal*	gl;

  _elements = new Dlist<SoakElement*>;
  
  GoalListIterator next_gl(gt->get_goals());
  while(gl = next_gl()) {
    
#ifdef DEBUG
    cerr << form("[Soak::Soak] (goal) \n");
    cerr << "\nGoal Name = " << gl->get_name() << "\n";
#endif

    if (!gl->generate_soak()) continue;

    KaTableBucketIterator next_ka(kt, gl);
    Binding* ka_binding;
      
    while(ka = next_ka(ka_binding)) {

#ifdef DEBUG
      cerr << form("[Soak::Soak] (ka) \n");
      cerr << "KA Name = " << ka->get_name() << "\n";
      cerr << "KA Binding) \n";
      ka_binding->print();
#endif

      instantiate(ka, ka_binding, gl);
    }

    // soak for new goal now generated, so goal is no longer new
    if (!test_generate)
      gl->clear_new();
  }

  if (!test_generate)
    wt->clear_new_all();
}


SoakElement* Soak::get_se(long nth)
{
  DlistIterator<SoakElement*>	next_se(_elements);
  SoakElement*			nth_element;

  if (nth <= _elements->get_count()-1) {
    nth_element = next_se();
    for (long i = 1;i <= nth; i++) {
      nth_element = next_se();
    }
    return nth_element;
  }
  return 0;
}


void Soak::operator=(Soak& s)
{
  if (this == &s) return;
  _elements->delete_entities();
  delete _elements;
  _elements = new Dlist<SoakElement*>;
  
  SoakIterator next_se(s.get_elements());
  SoakElement* se;

  se = next_se();
  while (se) {
    SoakElement*	se2 = new SoakElement;
    *se2 = *se;
    _elements->append(se2);
  }
}


void Soak::print(ostream& s)
{ 
  SoakIterator next_se(_elements);
  SoakElement* se;

  s << "Size:	" << _elements->get_count() << "\n";

  while(se = next_se()) {
    se->print(s);
    s << "\n";
  }
}


//

SoakElement* Soak::get_random()
{
  static Random_Uniform_Integer rui = random_uniform(_elements->get_count());
  return get_se(rui.generate());
}


SoakElement* Soak::get_priority_first()
{
  SoakElement*		se;
  SoakIterator		next_se(_elements);

  double max_priority = get_se(0)->eval_priority();
  long max_index = 0;

  double p;

  se = next_se();
  for (long i = 1; i < _elements->get_count(); i++)
    if ((p=next_se()->eval_priority()) > max_priority) {
      max_priority = p;
      max_index = i;
    }
    
  return get_se(max_index);
}


SoakElement* Soak::get_priority_random()
{
  SoakElement*		se;
  SoakIterator		next_se(_elements);

  long* index = new long[_elements->get_count()];
  index[0] = 0;
  long max_count = 0;

  double p;

  double max_priority = -MAXLONG;

  for (long i = 0; i < _elements->get_count(); i++)
    if ((p=next_se()->eval_priority()) > max_priority) {
      max_priority = p;
      max_count = 0;
      index[0] = i;
    }
    else if (p == max_priority) {
      index[++max_count] = i;
    }

  long max_index = index[random_uniform(max_count+1)];
  delete [] index;
  
  return get_se(max_index);
}


//
// Soak List
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
Soak* SoakList::get_soak(long level)
{
  DlistIterator<Soak*>	next_soak(soaks);
  Soak*			nth_soak;

  if (level <= soaks->get_count()) {
    nth_soak = next_soak();
    for (long i = 1;i < soaks->get_count(); i++) {
      nth_soak = next_soak();
    }
    return nth_soak;
  }
  return 0;
}


/////////////////////////////////////////////////////////////////////////////
Soak* SoakList::add(Soak* soak)
{
  soaks->append(soak);
  return soak;
}
