//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: wm.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/wm.h,v $
//  
//  File	      : wm.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : World Model 
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:16:22 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:03:06 1996
//  Update Count      : 238
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __WM_H__
#define __WM_H__

#include <string.h>
#include "relation.h"
#include "symtab.h"
#include "exp.h"

class Binding;

//
// Wolrd Model Table
//

#define WM_TAB_SIZE	 4096 /* Initial size to allocate for symbol table. */
#define WM_HASHTAB_SIZE	 1009 /* */
#define WM_TAB_INCREMENT 1024 /* Increment when symbol table must grow. */

//

class WmRelation : public Symbol
{
  Relation*	_relation;	// The content of the WM Relation Table
  int		_new_tag;	// set to 1 for new entry and cleared at every cycle

  void init(Relation* r, int n=1)
  {
    _relation = r;
    _new_tag = n;
  } 


public:

  ~WmRelation()
  {
    delete _relation;
  }

  WmRelation(const char* name, ExpList* args=0)
  : Symbol(name)
  {
    init(new Relation(name, args));
  }

  WmRelation(Relation* rel)
  // NOTE: The supplied relation will be DELETED when this WmRelation is deleted.
  : Symbol(rel->get_name())
  {
    init(rel);
  }

  WmRelation(WmRelation& wm_rel)
  : Symbol(wm_rel)
  {
    init(new Relation(wm_rel._relation));
  }

  void operator=(WmRelation& wm_rel)
  {
    Symbol::operator=(wm_rel);
    if (this == &wm_rel) return;
    delete _relation;
    init(new Relation(wm_rel._relation), wm_rel._new_tag);
  }

public:

  Relation*	get_relation() const	{ return _relation; }

  int		is_new() const		{ return _new_tag; }
  void		clear_new()		{ _new_tag = 0; }

public:

  int match_relation(const Relation* patt_relation, Binding* patt_binding);
};


//
//
//

class WmTable : public SymbolTable
{
// Relation is hased on the name of the relation

friend class WmTableIterator;
friend class WmTableBucketIterator;

public:

  WmTable(long wmtab_sz		= WM_TAB_SIZE, 
	  long wmtab_inc_sz	= WM_TAB_INCREMENT,
	  long hashtab_sz	= WM_HASHTAB_SIZE)
  : SymbolTable(wmtab_sz, wmtab_inc_sz, hashtab_sz)
  {};

public:

  WmRelation* 	lookup(long id) const
  {
    return (WmRelation*) SymbolTable::lookup(id);
  }
  
  WmRelation* 	lookup(const char* name) const
  {
    return (WmRelation*) SymbolTable::lookup(name);
  }

  //
  
  int	match(Relation* relation, Binding* binding);
	
  void 	assert(Relation* r, Binding* b=0);
  void 	retract(Relation* r, Binding* b=0);
  void 	update(Relation* old_rel, Relation* new_rel, Binding* b=0);
	
  int	is_new(long id) const	{ return lookup(id)->is_new(); }
  int	any_new();
  void	clear_new_all();

  void	print(ostream& s=cout) const;
};


class WmTableIterator
{
  WmTable*	_table;
  long 		_id;			
  long 		_max;

public:
  
  WmTableIterator(WmTable* wt, long start = 0)
  {
    _table = wt;
    _id = start;
    _max = _table->_nextid - 1;
  }

  WmRelation* operator()()
  {
    return (_id <= _max) ? _table->lookup(_id++) : 0; 
  }
};


class WmTableBucketIterator
{
  WmTable*		_table;
  const Relation*	_relation;
  SymbolListIterator* 	_bucket;
  
public:

  ~WmTableBucketIterator()
  {
    delete _bucket;
  }

  WmTableBucketIterator(WmTable* wt, const Relation* relation)
  {
    _table = wt;
    _relation = relation;
    _bucket = new SymbolListIterator(wt->get_bucket(_relation->get_name()));
  }

public:
  
  WmRelation* operator()(Binding* binding)
  {	
    WmRelation* wr = (WmRelation*) _bucket->operator()();

    if (!wr) return 0;

    if (wr->match_relation(_relation, binding))
      return wr;
    else
      return operator()(binding);
  }
  
  void remove_this()
  {
    _table->remove_this(_bucket);
  }
};

#endif
